import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;

/**
 * Represent a colored dot that moves around within the world according to rules
 * specified for the world, not for the dot.  The initial direction for the dot
 * may be specified by its color, or it may instead be random.
 * 
 * In this (unfininshed) version, the second dot created (id 1) is anointed the
 * bigDot and becomes the leader of a worm created by each of the dots following
 * the tail of the string of dots following bigDot. In some runs, not all dots
 * will become followers unless the user pauses the run and moves the leftovers
 * around a bit.
 * 
 * @author Marla Parker 
 * @version March 11, 2009
 */
public class Dot extends Mover
{
    /*
     * The different possible ROYGBIV colors for Dots.
     */
    private final static int RED = 0;
    private final static int ORANGE = 1;
    private final static int YELLOW = 2;
    private final static int GREEN = 3;
    private final static int BLUE = 4;
    private final static int INDIGO = 5;
    private final static int VIOLET = 6;
    
    private static final double WORLD_MARGIN = 2.0;    
    
    private static int nextId = 1;
    private static Dot bigDot;
    private ArrayList<Dot> worm;
    // right now only the work for bigDot is actually used, but maybe I will have
    // more worms in the future, so this field is not static
    
    private boolean follower = false;   // this dot is not a follower (yet)
    private int id;
    private int myColor;
    // this could be final, because right now the color of a specific dot does 
    // not change.  But I might want to change it later, so I'm not making the
    // myColor field final.  
        
    /**
     * Create a new dot with its own unique id.  Assign the second created dot,
     * id 1, to be the bigDot.
     * 
     * @param color The color for this dot
     * @param dirByColor Should the direction be random or specified by the color?
     */
    public Dot(int color, boolean dirByColor)
    {
        /*
         * Give each dot a unique id, so they can play
         * follow the leader later on.
         */
        id = nextId;
        nextId++;
        setColor(color);
        if (id == 1) {
            GreenfootImage img = getImage();
//            img.setColor(java.awt.Color.black);
//            img.fill();
            bigDot = this;
            worm = new ArrayList<Dot>();
        }
        /*
         * Random orientation, instead of all "facing" right to start with.  
         * More fun: make each color start with a particular orientation! 
         * What is 360/7 anyway?  51.4ish?
         */
        if (dirByColor) {
            setRotation((int) (color * 51.4));
        } else {
            setRotation((int) (Math.random() * 360));
        }
    }
    
    /**
     * resetNextId - reset the nextID counter in this class
     * Cannot remember why I decided to start at 1, nor why this is needed
     * at all in this version.... 
     */
    public static void resetNextId()
    {
        nextId = 1;
    }
    
    /**
     * Act - This method is called whenever the 'Act' or 'Run' button gets 
     * pressed in the environment.
     * 
     * If the dot is a follower, do nothing.  The bigDot act call will take care of its followers.
     * If it is not a follower, check the wall bounce before moving.
     * If it is bigDot, capture a nearby dot, if any, and make it a follower.  Make the followers into a
     * follow-the-leader worm of dots.
     */
    public void act() 
    {
        if (follower) {
            /*
             * If there was no other code to make this follower move, it would be
             * frozen now, like freeze tag.  If dirByColor == true and 
             * wallBounce == false, then predict what will the end state be if this 
             * scenario is left to run.  Set the speed very fast to let it stabalize,
             * then (if necessary) slow it down again to see what the end state is.
             */
            return;
        }
        
        // need to check the wall bounce whether it is the bigDot or not
        checkWallBounce();
        
        // variation: have the blue dots follow the mouse
        /* 
         * 
        if (getColor() == BLUE) {    
            
            MouseInfo mi = Greenfoot.getMouseInfo();
            if (mi != null) {
                setLocation(mi.getX(), mi.getY());
            }
         */
        if (this == bigDot) {           
            int xTail = bigDot.getX();
            int yTail = bigDot.getY();
            // move bigDot here
            move();
            // move all the followers
            for (Dot convert : worm) {
                int oldX = convert.getX();
                int oldY = convert.getY();
                convert.setLocation(xTail, yTail);
                xTail = oldX;
                yTail = oldY;
            }
            // Dot newConvert = (Dot) getOneObjectAtOffset(0, 0, Dot.class);
            // no this will always get the same object and get stuck
            // need to get all the objects and add one new follower
                /**
                 * Exercise: write a scenario that:
                 * a. shows the grid lines
                 * b. shows the outline of an image
                 * c. shows the center of the image
                 * d. shows the results of the various getObject... calls
                 * move objects around and then step with the Act button
                 * do some easy visual to show what objects are "gotten" like
                 * change their color
                 * e. then do this with various stock gf images
                 */
            List bumpedDots = getObjectsAtOffset(0, 0, Dot.class);
            boolean gotDot = false;
            Iterator<Dot> bd = bumpedDots.iterator();
            while (!gotDot && bd.hasNext()) {
                Dot newConvert = bd.next();
                if ((newConvert != bigDot) && !newConvert.follower) {
                    // we have a new follower
                    newConvert.follower = true;
                    gotDot = true;
                    newConvert.setRotation(getRotation());
                    newConvert.setLocation(xTail, yTail);
                    worm.add(newConvert);
                }
            }
        } else {
            // this is not bigDot, so just move normally
            move();
        }
    } 
    
    
    /**
     * setColor of this dot, both the image and the myColor variable
     * @param color One of the seven rainbow colors.
     */
    public void setColor(int color)
    {
        switch (color) {
            case RED: 
                setImage("r.png");
                break;
            case ORANGE: 
                setImage("o.png");
                break;
            case YELLOW: 
                setImage("y.png");
                break;
            case GREEN: 
                setImage("g.png");
                break;
            case BLUE: 
                setImage("b.png");
                break;
            case INDIGO: 
                setImage("i.png");
                break;
            case VIOLET: 
                setImage("v.png");
                break;
            default:
                System.err.printf("ooooooopps - what color is this?%d\n", color);
            }
        myColor = color;
    }
    
    /**
     * getColor of this dot
     * @return The color of this dot.
     */
    public int getColor()
    {
        return myColor;
    }
    
    
    /**
     * Determine if and how to deflect off one of the edges of the world. Set
     * the new rotation value in order to simulate a proper deflection bounce.
     * If we are not at the edge of the world,
     * then set the current heading (rotation) for no turn needed at all.
     */
    public void checkWallBounce()
    {
        int r = getRotation();
        int x = getX();
        int y = getY();

        PetriDish pd = (PetriDish) getWorld();
        boolean bounce = pd.getWallBounce();
        
        if(x < WORLD_MARGIN) {
            // hit the left edge
            if (bounce) {
                // change the direction of motion
                setRotation((180-r) % 360);
            } else {
                // change the location to the opposite side and keep going
                setLocation((int) (pd.getWidth() - WORLD_MARGIN), y);
            }
        } else if(x > pd.getWidth() - WORLD_MARGIN) {
            // hit the right edge
            if (bounce) {
                setRotation((180-r) % 360);
            } else {
                setLocation((int) WORLD_MARGIN, y);
            }
        } else if(y < WORLD_MARGIN) {
            // hit the top edge
            if (bounce) {
                setRotation(-r);
            } else {
                setLocation(x, (int) (pd.getHeight() - WORLD_MARGIN)); 
            }
        } else if(getY() > pd.getHeight() - WORLD_MARGIN) {            
            // hit the top edge
            if (bounce) {
                setRotation(-r);
            } else {
                setLocation(x, (int) WORLD_MARGIN);
            }
        } 
    }    

}
